# -*- coding:utf8 -*-

import numpy as np
import copy
import gensim
import json


def word2vec(model_path, file_path, saving_path):
    # model = gensim.models.KeyedVectors.load_word2vec_format(model_path, binary=True)
    model = gensim.models.Word2Vec.load(model_path)
    print("Successfully load word2vec model!")
    lines = np.array(map(lambda x: x.split(), open(file_path).readlines())).reshape(-1)
    print("The total word number is: %d" % lines.shape[0])
    embedding_size = model['a'].shape[0]
    UNK = 0.2 * np.random.random(size=embedding_size) - 0.1

    words = []
    words.append(np.zeros([embedding_size], dtype=np.float32))
    count = 0
    for i in range(lines.shape[0]):
        word = lines[i]
        if word.lower() in model:
            words.append(model[word.lower()])
        else:
            count += 1
            words.append(UNK)

    print("The number of 'UNK' word is: %d" % count)
    np.save(saving_path, words)


def genBatch(data):

    max_length = 0

    max_sentence_num = len(data[0])
    for doc in data:
        for sentence in doc:
            if len(sentence) > max_length:
                max_length = len(sentence)
        for i in xrange(max_sentence_num - len(doc)):
            doc.append([0])

    tmp = map(lambda doc: np.asarray(map(lambda sentence: sentence + [0] * (max_length - len(sentence)), doc),
                                     dtype=np.int32), data)
    tmp = reduce(lambda doc, docs: np.concatenate((doc, docs), axis=0), tmp)
    return tmp


def genLenBatch(lengths, maxsentencenum):
    lengths = map(lambda length: np.asarray(length + [1] * (maxsentencenum - len(length)), dtype=np.int32), lengths)
    return reduce(lambda x, y: np.concatenate((x, y), axis=0), lengths)


def gen_word_mask(docs_batch):
    mask = copy.deepcopy(docs_batch)
    mask = map(lambda x: map(lambda y: [1, 0][y == 0], x), mask)
    mask = np.asarray(mask, dtype=np.int32).T
    mask[0] = np.ones([mask.shape[1]], dtype=np.int32)
    return mask.T


def gen_docs_char(docs_batch, voc, char_voc):
    max_word_length = 10    # 10 is enough
    docs_char = copy.deepcopy(docs_batch)
    docs_char = map(lambda x: map(lambda y: map(lambda ch: char_voc.get_id(ch) + 1, voc.get_word(y - 1)), x), docs_char)
    docs_char = map(lambda x: map(lambda y: [y, y[:max_word_length - 1]][len(y) > max_word_length], x), docs_char)
    docs_char = map(lambda x: map(lambda y: y + [0] * (max_word_length - len(y)), x), docs_char)
    return docs_char


def gen_sentence_mask(sentence_num):
    max_num = sentence_num[0]
    mask = np.asarray(map(lambda num: [1] * num + [0] * (max_num - num), sentence_num), dtype=np.int32)
    return mask


def mySentences(filename):
    lines = map(lambda x: x.split('\t\t'), open(filename).readlines())
    docs = map(lambda x: x[3][0: len(x[3]) - 1], lines)
    docs = map(lambda x: x.split(' <sssss> '), docs)
    docs = map(lambda doc: map(lambda sentence: sentence.split(' '), doc), docs)
    sent = []
    for i in range(len(docs)):
        sent = sent + docs[i]
    return sent


def generate_word2vec():
    dev_sent = mySentences("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/dev.txt")
    print("Successfully load dev!")
    test_sent = mySentences("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/test.txt")
    print("Successfully load test!")
    train_sent = mySentences("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/train.txt")
    print("Successfully load train!")
    sent = dev_sent + test_sent + train_sent
    del dev_sent, test_sent, train_sent
    model = gensim.models.Word2Vec(sent, size=200, min_count=1)
    model.save("./data/IMDB/Word2VecModel")


class Dataset(object):
    def __init__(self, filename, voc=None, batch_size=32):

        lines = map(lambda x: x.split('\t\t'), open(filename).readlines())
        labels = np.array(
            map(lambda x: int(x[2]) - 1, lines),
            dtype=np.int32)
        index_offset = np.arange(labels.shape[0]) * 10   # 10 class
        labels_one_hot = np.zeros((labels.shape[0], 10))
        labels_one_hot.flat[index_offset + labels] = 1

        docs = map(lambda x: x[3][0: len(x[3]) - 1], lines)
        docs = map(lambda x: x.split(' <sssss> '), docs)
        docs = map(lambda doc: map(lambda sentence: sentence.split(' '), doc), docs)
        docs = map(lambda doc: map(
            lambda sentence: filter(lambda wordid: wordid != -1, map(lambda word: voc.get_id(word) + 1, sentence)), doc),
                   docs)

        tmp = zip(docs, labels_one_hot)
        tmp.sort(lambda x, y: len(y[0]) - len(x[0]))
        docs, labels_one_hot = zip(*tmp)

        sentence_num = map(lambda x: len(x), docs)
        length = map(lambda doc: map(lambda sentence: len(sentence), doc), docs)
        self._epoch = len(docs) // batch_size

        self._labels_one_hot = []
        self._docs = []
        # self._docs_char = []
        self._length = []
        self._word_mask = []
        self._sentence_num = []
        self._sentence_mask = []
        self._max_sentence_num = []
        
        for i in range(self._epoch):
            self._max_sentence_num.append(sentence_num[i * batch_size])
            self._length.append(genLenBatch(length[i * batch_size: (i + 1) * batch_size], sentence_num[i * batch_size]).tolist())
            doc_batch = genBatch(docs[i * batch_size: (i + 1) * batch_size])
            self.docs.append(doc_batch.tolist())
            self._labels_one_hot.append(np.array(labels_one_hot[i * batch_size: (i + 1) * batch_size], dtype=np.int32).tolist())
            self._sentence_num.append(np.asarray(sentence_num[i * batch_size: (i + 1) * batch_size], dtype=np.int32).tolist())
            # self._docs_char.append(gen_docs_char(doc_batch, voc, char_voc))
            self._word_mask.append(gen_word_mask(doc_batch).tolist())
            self._sentence_mask.append(gen_sentence_mask(sentence_num[i * batch_size: (i + 1) * batch_size]).tolist())

    @property
    def labels_one_hot(self):
        return self._labels_one_hot

    @property
    def docs(self):
        return self._docs

    @property
    def epoch(self):
        return self._epoch

    @property
    def word_mask(self):
        return self._word_mask

    @property
    def sentence_mask(self):
        return self._sentence_mask

    @property
    def length(self):
        return self._length

    @property
    def sentence_num(self):
        return self._sentence_num

    """
    @property
    def docs_char(self):
        return self._docs_char
    """
    def save(self, path):
        f = open(path, "wb")
        dict_t = {"docs": self.docs,
                  "labels_one_hot": self.labels_one_hot,
                  "length": self.length,
                  "word_mask": self.word_mask,
                  "sentence_num": self.sentence_num,
                  "sentence_mask": self.sentence_mask}
        json.dump(dict_t, f)


class Charlist(object):

    def __init__(self, filename):
        line = open(filename).readline()
        self._size = len(line)
        print("Charlist size is: %d" % self._size)
        line = [item for item in line]

        voc = [(item[0], item[1]) for item in zip(line, xrange(self._size))]
        self._voc = dict(voc)

    def get_id(self, char):
        if char in self._voc:
            return self._voc[char]
        else:
            return -1

    @property
    def size(self):
        return self._size


class Wordlist(object):
    def __init__(self, filename):
        lines = map(lambda x: x.split(), open(filename).readlines())
        self._size = len(lines)
        print("wordlist size is: %d" % self._size)

        voc = [(item[0][0], item[1]) for item in zip(lines, xrange(self._size))]
        self._voc = dict(voc)

        voc_v = [(item[1], item[0][0]) for item in zip(lines, xrange(self._size))]
        self._voc_v = dict(voc_v)

    def get_id(self, word):
        if word in self._voc:
            return self._voc[word]
        else:
            return -1

    def get_word(self, word_id):
        if word_id in self._voc_v:
            return [item for item in self._voc_v[word_id]]
        else:
            return []

    @property
    def size(self):
        return self._size


def pre_processing():
    voc = Wordlist("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/wordlist.txt")
    dev_data = Dataset("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/dev.txt", voc)
    dev_data.save("./data/IMDB/val_32.json")
    print("dev data is done!")
    del dev_data
    train_data = Dataset("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/train.txt", voc)
    train_data.save("./data/IMDB/train_32.json")
    print("train data is done!")
    del train_data
    test_data = Dataset("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/test.txt", voc)
    test_data.save("./data/IMDB/test_32.json")
    print("test data is done!")
    del test_data

# word2vec("./data/IMDB/Word2VecModel",
#          "/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/wordlist.txt",
#          "./data/IMDB/word_embedding")


pre_processing()

# generate_word2vec()

"""
voc = Wordlist("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/wordlist.txt")
data = Dataset("/media/rowitzou/WORK/NLP/sentiment-classification/data/IMDB/dev.txt", voc)
# print(len(data.labels[0]))
print(data.epoch)
for i in range(len(data.docs)):
    print(i, np.array(data.docs[i]).shape)
print(np.array(data.docs[260]))
print(np.array(data.word_mask[260]))
print(np.array(data.sentence_mask[260]))
print(np.array(data.length[260]))
print(np.array(data.sentence_num[260]))
print(np.array(data.labels_one_hot)[260])
"""
