# -*- coding:utf8 -*-

import tensorflow as tf


class MeanPoolLayer(object):

    def __init__(self, input, config, mask, is_bi=False, name=None):
        hidden_size = config.hidden_size
        if is_bi:
            if name == "sent":
                hidden_size *= 2
            else:
                hidden_size *= 4

        mask_extend = tf.tile(tf.expand_dims(mask, 2), [1, 1, hidden_size])
        sum = tf.reduce_sum(mask_extend * input, axis=1)
        self._output = sum / tf.tile(tf.expand_dims(tf.reduce_sum(mask, axis=1), 1), [1, hidden_size])

    @property
    def output(self):
        return self._output


"""
class StructuredWordLayer(object):

    def __init__(self, input, config, mask, is_bi=False):

        hidden_size = config.hidden_size
        if is_bi:
            hidden_size *= 2
        axis_1 = tf.shape(input)[0]
        axis_2 = tf.shape(input)[1]
        Wc = tf.get_variable("structure_wc", [hidden_size, hidden_size], dtype=tf.float32)
        Wp = tf.get_variable("structure_wp", [hidden_size, hidden_size], dtype=tf.float32)
        Wa = tf.get_variable("structure_wa", [hidden_size, hidden_size], dtype=tf.float32)
        Wr = tf.get_variable("structure_wr", [hidden_size, 1], dtype=tf.float32)

        word_valid_count = tf.reduce_sum(mask, axis=1)
        mask_extend = tf.tile(tf.expand_dims(mask, 2), [1, 1, hidden_size])
        Tc = tf.tanh(tf.matmul(tf.reshape(input * mask_extend, [-1, hidden_size]), Wc))
        Tp = tf.tanh(tf.matmul(tf.reshape(input * mask_extend, [-1, hidden_size]), Wp))
        Fr = tf.matmul(tf.reshape(input * mask_extend, [-1, hidden_size]), Wr)

        Tc = tf.reshape(Tc, [axis_1, axis_2, hidden_size]) * mask_extend
        Tp_Wa = tf.reshape(tf.matmul(Tp, Wa), [axis_1, axis_2, hidden_size]) * mask_extend
        Fr = tf.reshape(Fr, [axis_1, axis_2]) * mask

        F = tf.matmul(tf.reshape(tf.tile(Tp_Wa, [1, 1, axis_2]), [-1, hidden_size]),
                      tf.transpose(tf.reshape(tf.tile(Tc, [1, axis_2, 1]), [-1, hidden_size])))
"""


class SupervisedWordAttentionLayer(object):

    def __init__(self, input_fw, input_bw, config, mask, softmax_mask, is_bi=False):

        hidden_size = config.hidden_size
        if is_bi:
            hidden_size *= 2
        axis_1 = tf.shape(input_fw)[0]
        axis_2 = tf.shape(input_fw)[1]
        Wf = tf.get_variable("word_f_w", [config.hidden_size, config.hidden_size], dtype=tf.float32)
        Wb = tf.get_variable("word_b_w", [config.hidden_size, config.hidden_size], dtype=tf.float32)
        # Wc = tf.get_variable("char_w", [config.char_embed_size, config.char_embed_size], dtype=tf.float32)
        b = tf.get_variable("word_b", [config.hidden_size], dtype=tf.float32)
        u_w = tf.get_variable("word_context", [config.hidden_size, 1])
        # bc = tf.get_variable("char_b", [config.char_embed_size], dtype=tf.float32)

        mask_extend = tf.tile(tf.expand_dims(mask, 2), [1, 1, config.hidden_size])
        input_fw_mask = tf.reshape(input_fw * mask_extend, [-1, config.hidden_size])
        input_bw_mask = tf.reshape(input_bw * mask_extend, [-1, config.hidden_size])
        u_hidden = tf.tanh(tf.matmul(input_fw_mask, Wf) + tf.matmul(input_bw_mask, Wb) + b)
        u_hidden = tf.reshape(tf.reshape(u_hidden, [axis_1, axis_2, config.hidden_size])
                              * mask_extend, [-1, config.hidden_size])
        u_hidden = (tf.reshape(tf.matmul(u_hidden, u_w), [axis_1, axis_2]) * mask) + softmax_mask
        weight = tf.nn.softmax(u_hidden)
        weight_extend = tf.tile(tf.expand_dims(weight, 2), [1, 1, hidden_size])
        self._output = tf.reduce_sum(weight_extend * tf.concat([input_fw, input_bw], 2), axis=1)
        self._word_degree = tf.sigmoid(u_hidden)

        self._saver = tf.train.Saver([Wf, Wb, b, u_w])

    @property
    def output(self):
        return self._output

    @property
    def word_degree(self):
        return self._word_degree

    @property
    def saver(self):
        return self._saver


class SupervisedSentAttentionLayer(object):

    def __init__(self, input_fw, input_bw, config, mask, softmax_mask, is_bi=False):
        hidden_size = config.hidden_size
        if is_bi:
            hidden_size *= 2
        axis_1 = tf.shape(input_fw)[0]
        axis_2 = tf.shape(input_fw)[1]
        Wf = tf.get_variable("sent_f_w", [hidden_size, hidden_size], dtype=tf.float32)
        Wb = tf.get_variable("sent_b_w", [hidden_size, hidden_size], dtype=tf.float32)
        # Wc = tf.get_variable("char_w", [config.char_embed_size, config.char_embed_size], dtype=tf.float32)
        b = tf.get_variable("sent_b", [hidden_size], dtype=tf.float32)
        u_w = tf.get_variable("sent_context", [hidden_size, 1])
        # bc = tf.get_variable("char_b", [config.char_embed_size], dtype=tf.float32)

        mask_extend = tf.tile(tf.expand_dims(mask, 2), [1, 1, hidden_size])
        input_fw_mask = tf.reshape(input_fw * mask_extend, [-1, hidden_size])
        input_bw_mask = tf.reshape(input_bw * mask_extend, [-1, hidden_size])
        u_hidden = tf.tanh(tf.matmul(input_fw_mask, Wf) + tf.matmul(input_bw_mask, Wb) + b)
        u_hidden = tf.reshape(tf.reshape(u_hidden, [axis_1, axis_2, hidden_size])
                              * mask_extend, [-1, hidden_size])
        u_hidden = (tf.reshape(tf.matmul(u_hidden, u_w), [axis_1, axis_2]) * mask) + softmax_mask
        weight = tf.nn.softmax(u_hidden)
        weight_extend = tf.tile(tf.expand_dims(weight, 2), [1, 1, hidden_size * 2])
        self._output = tf.reduce_sum(weight_extend * tf.concat([input_fw, input_bw], 2), axis=1)
        self._sent_degree = tf.sigmoid(u_hidden)

        self._saver = tf.train.Saver([Wf, Wb, b, u_w])

    @property
    def output(self):
        return self._output

    @property
    def sent_degree(self):
        return self._sent_degree

    @property
    def saver(self):
        return self._saver


class LocalWordAttentionLayer(object):

    def __init__(self, input, config, mask, softmax_mask, is_bi=False):

        hidden_size = config.hidden_size
        if is_bi:
            hidden_size *= 2
        axis_1 = tf.shape(input)[0]
        axis_2 = tf.shape(input)[1]
        W = tf.get_variable("word_w", [hidden_size, hidden_size], dtype=tf.float32)
        b = tf.get_variable("word_b", [hidden_size], dtype=tf.float32)
        u_w = tf.get_variable("word_context", [hidden_size, 1], dtype=tf.float32)

        mask_extend = tf.tile(tf.expand_dims(mask, 2), [1, 1, hidden_size])
        # mask_extend_2 = tf.tile(tf.expand_dims(mask, 2), [1, 1, hidden_size])
        u_hidden = tf.tanh(tf.matmul(tf.reshape(input * mask_extend, [-1, hidden_size]), W) + b)
        u_hidden = tf.reshape(tf.reshape(u_hidden, [axis_1, axis_2, hidden_size])
                              * mask_extend, [-1, hidden_size])
        u_hidden = (tf.reshape(tf.matmul(u_hidden, u_w), [axis_1, axis_2]) * mask) + softmax_mask
        self._weight = weight = tf.nn.softmax(u_hidden)
        weight_extend = tf.tile(tf.expand_dims(weight, 2), [1, 1, hidden_size])
        self._output = tf.reduce_sum(weight_extend * input, axis=1)
        self._word_degree = tf.sigmoid(u_hidden)
        self._saver = tf.train.Saver([W, b, u_w])

    @property
    def output(self):
        return self._output

    @property
    def weight(self):
        return self._weight

    @property
    def saver(self):
        return self._saver

    @property
    def word_degree(self):
        return self._word_degree


class LocalSentAttentionLayer(object):

    def __init__(self, input, config, mask, softmax_mask, is_bi=False):

        hidden_size = config.hidden_size
        if is_bi:
            hidden_size *= 4
        axis_1 = tf.shape(input)[0]
        axis_2 = tf.shape(input)[1]
        W = tf.get_variable("sent_w", [hidden_size, hidden_size], dtype=tf.float32)
        b = tf.get_variable("sent_b", [hidden_size], dtype=tf.float32)
        u_s = tf.get_variable("sent_context", [hidden_size, 1], dtype=tf.float32)

        mask_extend = tf.tile(tf.expand_dims(mask, 2), [1, 1, hidden_size])
        # mask_extend_2 = tf.tile(tf.expand_dims(mask, 2), [1, 1, config.hidden_size])
        u_hidden = tf.tanh(tf.matmul(tf.reshape(input * mask_extend, [-1, hidden_size]), W) + b)
        u_hidden = tf.reshape(tf.reshape(u_hidden, [axis_1, axis_2, hidden_size])
                              * mask_extend, [-1, hidden_size])
        u_hidden = (tf.reshape(tf.matmul(u_hidden, u_s), [axis_1, axis_2]) * mask) + softmax_mask
        self._weight = weight = tf.nn.softmax(u_hidden)
        weight_extend = tf.tile(tf.expand_dims(weight, 2), [1, 1, hidden_size])
        self._output = tf.reduce_sum(weight_extend * input, axis=1)
        self._sent_degree = tf.sigmoid(u_hidden)
        self._saver = tf.train.Saver([W, b, u_s])

    @property
    def saver(self):
        return self._saver

    @property
    def output(self):
        return self._output

    @property
    def weight(self):
        return self._weight

    @property
    def sent_degree(self):
        return self._sent_degree
