# -*- coding:utf8 -*-

import tensorflow as tf


class InputHiddenLayer(object):

    def __init__(self, input, config):
        axis_1 = tf.shape(input)[0]
        axis_2 = tf.shape(input)[1]
        W = tf.get_variable("input_W", [config.embedding_size, config.hidden_size], dtype=tf.float32)
        b = tf.get_variable("input_b", [config.hidden_size], dtype=tf.float32)
        output = tf.matmul(tf.reshape(input, [-1, config.embedding_size]), W) + b
        self._output = tf.reshape(output, [axis_1, axis_2, config.hidden_size])

        self._saver = tf.train.Saver([W, b])

    @property
    def output(self):
        return self._output

    @property
    def saver(self):
        return self._saver


class SoftmaxLayer(object):

    def __init__(self, input, config, is_bi=False):
        hidden_size = config.hidden_size
        if is_bi:
            hidden_size *= 4
        W = tf.get_variable("output_W", [hidden_size, config.class_size], dtype=tf.float32)
        b = tf.get_variable("output_b", [config.class_size], dtype=tf.float32)
        self._hidden_output = tf.matmul(input, W) + b
        self._output = tf.nn.softmax(self._hidden_output)

        self._saver = tf.train.Saver([W, b])

    @property
    def output(self):
        return self._output

    @property
    def hidden_output(self):
        return self._hidden_output

    @property
    def saver(self):
        return self._saver
