# -*- coding:utf8 -*-

import tensorflow as tf


class LSTMLayer(object):

    def __init__(self, inputs, config, is_training, sequence_length, name):

        cell = tf.nn.rnn_cell.LSTMCell(num_units=config.hidden_size)

        if is_training and config.keep_prob < 1:
            cell = tf.nn.rnn_cell.DropoutWrapper(cell, output_keep_prob=config.keep_prob)

        self._saver = tf.train.Saver()

        # self._initial_state = cell.zero_state(batch_size=batch_size, dtype=tf.float32)

        self._output, self._final_state = tf.nn.dynamic_rnn(cell, inputs,
                                                            dtype=tf.float32,
                                                            sequence_length=sequence_length,
                                                            scope=name)

    """
    @property
    def initial_state(self):
        return self._initial_state
    """
    @property
    def final_state(self):
        return self._final_state

    @property
    def output(self):
        return self._output

    @property
    def saver(self):
        return self._saver


class Bi_LSTMLayer(object):

    def __init__(self, inputs, config, is_training, sequence_length, name):

        hidden_size = config.hidden_size
        if name == "doc":
            hidden_size *= 2

        fw_cell = tf.nn.rnn_cell.LSTMCell(num_units=hidden_size)

        if is_training and config.keep_prob < 1:
            fw_cell = tf.nn.rnn_cell.DropoutWrapper(fw_cell, output_keep_prob=config.keep_prob)

        bw_cell = tf.nn.rnn_cell.LSTMCell(num_units=hidden_size)

        if is_training and config.keep_prob < 1:
            bw_cell = tf.nn.rnn_cell.DropoutWrapper(bw_cell, output_keep_prob=config.keep_prob)

        self._saver = tf.train.Saver()

        # self._initial_state = cell.zero_state(batch_size=batch_size, dtype=tf.float32)

        (output_fw, output_bw), self._final_state = tf.nn.bidirectional_dynamic_rnn(fw_cell, bw_cell, inputs,
                                                                                    dtype=tf.float32,
                                                                                    sequence_length=sequence_length,
                                                                                    scope=name)
        self._output_concat = tf.concat([output_fw, output_bw], 2)
        self._output_fw = output_fw
        self._output_bw = output_bw

    @property
    def saver(self):
        return self._saver

    @property
    def final_state(self):
        return self._final_state

    @property
    def output_concat(self):
        return self._output_concat

    @property
    def output_fw(self):
        return self._output_fw

    @property
    def output_bw(self):
        return self._output_bw

