# -*- coding:utf8 -*-

from EmbLayer import EmbLayer
from LSTMLayer import LSTMLayer, Bi_LSTMLayer
from HiddenLayer import SoftmaxLayer, InputHiddenLayer
from AttentionLayer import *
import tensorflow as tf


class LSTMModel(object):

    def __init__(self, config, is_training):

        self._input = tf.placeholder(tf.int32, [None, None])    # sent_num * sent_length
        self._label = tf.placeholder(tf.float32, [config.batch_size, config.class_size])    # doc_num * class_num
        self._input_char = tf.placeholder(tf.int32, [None, None, config.max_word_length])
        self._word_sequence_length = tf.placeholder(tf.int32, [None])   # sent_num
        self._sent_sequence_length = tf.placeholder(tf.int32, [config.batch_size])      # doc_num
        self._word_mask = tf.placeholder(tf.float32, [None, None])    # sent_num * sent_length
        self._word_softmax_mask = tf.placeholder(tf.float32, [None, None])
        self._sent_mask = tf.placeholder(tf.float32, [config.batch_size, None])   # doc_num * sent_num
        self._sent_softmax_mask = tf.placeholder(tf.float32, [config.batch_size, None])

        self._supervised_attention_word = tf.placeholder(tf.float32, [None, None])
        self._supervised_attention_sent = tf.placeholder(tf.float32, [config.batch_size, None])

        emb_layer = EmbLayer(self._input, config)

        input_hidden_layer = InputHiddenLayer(emb_layer.output, config)

        self._input_hidden_layer_saver = input_hidden_layer.saver

        # char_cnn_layer = CharCNNLayer(self._input_char, config)
        # self._char_cnn_layer_output = char_cnn_layer.output

        # sent_lstm_layer = LSTMLayer(emb_layer.output, config, is_training, self._word_sequence_length, "sent")
        sent_lstm_layer = Bi_LSTMLayer(input_hidden_layer.output, config, is_training, self._word_sequence_length, "sent")

        self._sent_lstm_saver = sent_lstm_layer.saver

        # sent_mean_pool_layer = MeanPoolLayer(sent_lstm_layer.output, config, self._word_mask)
        # sent_mean_pool_layer = MeanPoolLayer(sent_lstm_layer.output, config, self._word_mask, is_bi=True, name="sent")
        # sent_attention_layer = LocalWordAttentionLayer(sent_lstm_layer.output, config,
        #                                                self._word_mask, self._word_softmax_mask, is_bi=False)
        sent_attention_layer = SupervisedWordAttentionLayer(sent_lstm_layer.output_fw,
                                                            sent_lstm_layer.output_bw,
                                                            config, self._word_mask,
                                                            self._word_softmax_mask,
                                                            is_bi=True)

        self._sent_attention_saver = sent_attention_layer.saver

        # doc_lstm_layer = LSTMLayer(tf.reshape(sent_attention_layer.output, [config.batch_size, -1, config.hidden_size]),
        #                            config, is_training, self._sent_sequence_length, "doc")
        doc_lstm_layer = Bi_LSTMLayer(tf.reshape(sent_attention_layer.output, [config.batch_size, -1, config.hidden_size * 2]),
                                      config, is_training, self._sent_sequence_length, "doc")

        self._doc_lstm_saver = doc_lstm_layer.saver

        # doc_mean_pool_layer = MeanPoolLayer(doc_lstm_layer.output, config, self._sent_mask)
        # doc_mean_pool_layer = MeanPoolLayer(doc_lstm_layer.output, config, self._sent_mask, is_bi=True, name="doc")
        # doc_attention_layer = LocalSentAttentionLayer(doc_lstm_layer.output, config,
        #                                               self._sent_mask, self._sent_softmax_mask, is_bi=False)
        doc_attention_layer = SupervisedSentAttentionLayer(doc_lstm_layer.output_fw,
                                                           doc_lstm_layer.output_bw,
                                                           config, self._sent_mask,
                                                           self._sent_softmax_mask,
                                                           is_bi=True)

        self._doc_attention_saver = doc_attention_layer.saver

        # pre = SoftmaxLayer(doc_mean_pool_layer.output, config)
        pre = SoftmaxLayer(doc_attention_layer.output, config, is_bi=True)

        self._output_hidden_layer_saver = pre.saver

        correct_pre = tf.equal(tf.argmax(tf.reshape(self._label, [-1, config.class_size]),
                                         axis=1), tf.argmax(pre.output, axis=1))

        self._mse = tf.losses.mean_squared_error(tf.argmax(tf.reshape(self._label, [-1, config.class_size]), axis=1),
                                                 tf.argmax(pre.output, axis=1))
        self._acc = tf.reduce_mean(tf.cast(correct_pre, tf.float32))
        """
        self._word_attention_loss = word_attention_loss = \
            tf.reduce_sum(0.5 * ((sent_attention_layer.weight - self._supervised_attention_word) ** 2))
        self._sent_attention_loss = sent_attention_loss = \
            tf.reduce_sum(0.5 * ((doc_attention_layer.weight - self._supervised_attention_sent) ** 2))
        """
        self._word_attention_loss = word_attention_loss = \
            - tf.reduce_sum(sent_attention_layer.word_degree * tf.log(self._supervised_attention_word + 0.00000001))
        self._sent_attention_loss = sent_attention_loss = \
            - tf.reduce_sum(doc_attention_layer.sent_degree * tf.log(self._supervised_attention_sent + 0.00000001))

        self._classification_loss = classification_loss = \
            - tf.reduce_sum(tf.reshape(self._label, [-1, config.class_size]) * tf.log(pre.output + 0.00000001))

        self._loss = loss = classification_loss \
                            + config.word_attention_loss_lambda * word_attention_loss \
                            + config.sent_attention_loss_lambda * sent_attention_loss

        if not is_training:
            return

        self._train_op = tf.train.AdamOptimizer().minimize(loss)

    @property
    def accuracy(self):
        return self._acc

    @property
    def train_op(self):
        return self._train_op

    @property
    def input(self):
        return self._input

    @property
    def input_char(self):
        return self._input_char

    @property
    def label(self):
        return self._label

    @property
    def word_sequence_length(self):
        return self._word_sequence_length

    @property
    def sent_sequence_length(self):
        return self._sent_sequence_length

    @property
    def supervised_attention_word(self):
        return self._supervised_attention_word

    @property
    def supervised_attention_sent(self):
        return self._supervised_attention_sent

    @property
    def word_mask(self):
        return self._word_mask

    @property
    def sent_mask(self):
        return self._sent_mask

    @property
    def word_soft_max(self):
        return self._word_softmax_mask

    @property
    def sent_soft_max(self):
        return self._sent_softmax_mask

    @property
    def loss(self):
        return self._loss

    @property
    def mse(self):
        return self._mse

    @property
    def input_hidden_layer_saver(self):
        return self._input_hidden_layer_saver

    @property
    def sent_lstm_saver(self):
        return self._sent_lstm_saver

    @property
    def sent_attention_saver(self):
        return self._sent_attention_saver

    @property
    def doc_lstm_saver(self):
        return self._doc_lstm_saver

    @property
    def doc_attention_saver(self):
        return self._doc_attention_saver

    @property
    def output_hidden_layer_saver(self):
        return self._output_hidden_layer_saver

    @property
    def classification_loss(self):
        return self._classification_loss

    @property
    def word_attention_loss(self):
        return self._word_attention_loss

    @property
    def sent_attention_loss(self):
        return self._sent_attention_loss
